from typing import Union, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message
from telebox.bot.consts import message_entity_types


class HashtagFilterCache(AbstractEventFilterCache):

    def create(self, event: Union[Message, MediaGroup]) -> set[str]:
        hashtags = set()
        messages = event.messages if isinstance(event, MediaGroup) else [event]

        for message in messages:
            for entity in message.get_entities():
                if entity.type == message_entity_types.HASHTAG:
                    hashtags.add(
                        message.get_entity_text(entity)
                    )

        return hashtags


class HashtagFilter(AbstractEventFilter):

    def __init__(self, hashtags: Iterable[str], ignore_case: bool, cache: HashtagFilterCache):
        self._hashtags = set()

        for i in hashtags:
            if not i.startswith("#"):
                i = f"#{i}"

            if ignore_case:
                i = i.lower()

            self._hashtags.add(i)

        self._ignore_case = ignore_case
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP
        }

    def check_event(self, event: Union[Message, MediaGroup]) -> bool:
        hashtags = self._cache.get(event)

        if not self._hashtags:
            return bool(hashtags)

        for i in hashtags:
            if self._ignore_case:
                i = i.lower()

            if i in self._hashtags:
                return True

        return False


class HashtagFilterFactory(AbstractEventFilterFactory):

    def __init__(self):
        self._cache = HashtagFilterCache()

    def get_filter(self, *hashtags: str, ignore_case: bool = True) -> HashtagFilter:
        return HashtagFilter(hashtags, ignore_case, self._cache)



from __future__ import annotations
from typing import TYPE_CHECKING

from telebox.dispatcher.filters.errors.filter import AbstractErrorFilter
if TYPE_CHECKING:
    from telebox.dispatcher.dispatcher import Event


class NoneErrorFilter(AbstractErrorFilter):

    def check_error(self, error: Exception, event: Event) -> bool:
        return True



from __future__ import annotations
from typing import Type, TYPE_CHECKING

from telebox.dispatcher.filters.errors.filter import AbstractErrorFilter
if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event


class TypeErrorFilter(AbstractErrorFilter):

    def __init__(self, *types: Type[Exception], strictly: bool = False):
        if not types:
            raise ValueError("No types!")

        self._types = types
        self._strictly = strictly

    def check_error(self, error: Exception, event: Event) -> bool:
        return type(error) in self._types if self._strictly else isinstance(error, self._types)



from dataclasses import dataclass

from telebox.utils.env.errors import EnvError


@dataclass
class Env:

    def initialize(self, **kwargs) -> None:
        for name, value in kwargs.items():
            if name not in vars(self):
                raise EnvError(f"No {name!r} field!")

            setattr(self, name, value)

        for name, value in vars(self).items():
            if value is None:
                raise EnvError(f"{name!r} is not set!")



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.consts import inline_query_result_types
from telebox.bot.types.types.inline_keyboard_markup import InlineKeyboardMarkup


@dataclass
class InlineQueryResultGame(Type):
    id: str
    game_short_name: str
    reply_markup: Optional[InlineKeyboardMarkup] = None
    type: str = inline_query_result_types.GAME



from typing import Any, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.utils.callback_data_builders.builder import AbstractCallbackDataBuilder
from telebox.bot.types.types.callback_query import CallbackQuery


class CallbackKeyFilterCache(AbstractEventFilterCache):

    def __init__(self, builder: AbstractCallbackDataBuilder):
        self._builder = builder

    def create(self, event: CallbackQuery) -> Any:
        if event.data is not None:
            return self._builder.get_key(event.data)


class CallbackKeyFilter(AbstractEventFilter):

    def __init__(self, keys: Iterable[Any], cache: CallbackKeyFilterCache):
        self._keys = set(keys)
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {EventType.CALLBACK_QUERY}

    def check_event(self, event: CallbackQuery) -> bool:
        key = self._cache.get(event)

        if self._keys:
            return key in self._keys

        return key is not None


class CallbackKeyFilterFactory(AbstractEventFilterFactory):

    def __init__(self, builder: AbstractCallbackDataBuilder):
        self._cache = CallbackKeyFilterCache(builder)

    def get_filter(self, *keys: Any) -> CallbackKeyFilter:
        return CallbackKeyFilter(keys, self._cache)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.chat import Chat
from telebox.bot.types.types.user import User
from telebox.bot.types.types.chat_member import ChatMember
from telebox.bot.types.types.chat_invite_link import ChatInviteLink


@dataclass
class ChatMemberUpdated(Type):
    chat: Chat
    from_: User
    date: datetime
    old_chat_member: ChatMember
    new_chat_member: ChatMember
    invite_link: Optional[ChatInviteLink] = None

    @property
    def chat_type(self) -> str:
        return self.chat.type

    @property
    def chat_id(self) -> int:
        return self.chat.id

    @property
    def user_id(self) -> int:
        return self.from_.id



from __future__ import annotations
from typing import TYPE_CHECKING

from telebox.dispatcher.filters.errors.filter import AbstractErrorFilter
if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event


class TextErrorFilter(AbstractErrorFilter):

    def __init__(self, *texts: str, full_match: bool = True, ignore_case: bool = False):
        if not texts:
            raise ValueError("No texts!")

        self._texts = {i.lower() for i in texts} if ignore_case else set(texts)
        self._full_match = full_match
        self._ignore_case = ignore_case

    def check_error(self, error: Exception, event: Event) -> bool:
        text = str(error)

        if self._ignore_case:
            text = text.lower()

        if self._full_match:
            return text in self._texts
        else:
            return any(i in text for i in self._texts)



from abc import ABC, abstractmethod


class AbstractCallbackDataBuilder(ABC):

    @abstractmethod
    def get_builded_data(self, key, value=None) -> str:
        pass

    @abstractmethod
    def get_parsed_data(self, data: str):
        pass

    def get_key(self, data: str):
        key, _ = self.get_parsed_data(data)

        return key

    def get_value(self, data: str):
        _, value = self.get_parsed_data(data)

        return value



from __future__ import annotations
from typing import Any, Optional, Iterable, TYPE_CHECKING

if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event
from telebox.dispatcher.errors import InvalidEventError
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message
from telebox.bot.types.types.inline_query import InlineQuery
from telebox.bot.types.types.chosen_inline_result import ChosenInlineResult
from telebox.bot.types.types.callback_query import CallbackQuery
from telebox.bot.types.types.shipping_query import ShippingQuery
from telebox.bot.types.types.pre_checkout_query import PreCheckoutQuery
from telebox.bot.types.types.poll import Poll
from telebox.bot.types.types.poll_answer import PollAnswer
from telebox.bot.types.types.chat_member_updated import ChatMemberUpdated
from telebox.bot.types.types.chat_join_request import ChatJoinRequest


TYPES_WITH_CALLBACK_QUERY_ID = frozenset({
    CallbackQuery
})
TYPES_WITH_INLINE_QUERY_ID = frozenset({
    InlineQuery
})
TYPES_WITH_SHIPPING_QUERY_ID = frozenset({
    ShippingQuery
})
TYPES_PRE_CHECKOUT_QUERY_ID = frozenset({
    PreCheckoutQuery
})


def get_event_chat_id(event: Event, *, strictly: bool = False) -> Optional[int]:
    return _get_event_attribute(
        attribute="chat_id",
        name="chat identifier",
        event=event,
        types=frozenset((
            Message,
            MediaGroup,
            CallbackQuery,
            ChatMemberUpdated,
            ChatJoinRequest
        )),
        strictly=strictly
    )


def get_event_user_id(event: Event, *, strictly: bool = False) -> Optional[int]:
    return _get_event_attribute(
        attribute="user_id",
        name="user identifier",
        event=event,
        types=frozenset((
            Message,
            MediaGroup,
            InlineQuery,
            ChosenInlineResult,
            CallbackQuery,
            ShippingQuery,
            PreCheckoutQuery,
            Poll,
            PollAnswer,
            ChatMemberUpdated,
            ChatJoinRequest
        )),
        strictly=strictly
    )


def get_event_message_topic_id(event: Event, *, strictly: bool = False) -> Optional[int]:
    return _get_event_attribute(
        attribute="message_topic_id",
        name="message topic identifier",
        event=event,
        types=frozenset((Message, MediaGroup)),
        strictly=strictly
    )


def get_event_sender_chat_id(event: Event, *, strictly: bool = False) -> Optional[int]:
    return _get_event_attribute(
        attribute="sender_chat_id",
        name="sender chat identifier",
        event=event,
        types=frozenset((Message, MediaGroup)),
        strictly=strictly
    )


def get_event_message_id(event: Event, *, strictly: bool = False) -> Optional[int]:
    return _get_event_attribute(
        attribute="message_id",
        name="message identifier",
        event=event,
        types=frozenset((Message, MediaGroup, CallbackQuery)),
        strictly=strictly
    )


def get_event_callback_query_id(event: Event, *, strictly: bool = False) -> Optional[str]:
    return _get_event_attribute(
        attribute="id",
        name="callback query identifier",
        event=event,
        types=(CallbackQuery,),
        strictly=strictly
    )


def get_event_inline_query_id(event: Event, *, strictly: bool = False) -> Optional[str]:
    return _get_event_attribute(
        attribute="id",
        name="inline query identifier",
        event=event,
        types=(InlineQuery,),
        strictly=strictly
    )


def get_event_shipping_query_id(event: Event, *, strictly: bool = False) -> Optional[str]:
    return _get_event_attribute(
        attribute="id",
        name="shipping query identifier",
        event=event,
        types=(ShippingQuery,),
        strictly=strictly
    )


def get_event_pre_checkout_query_id(event: Event, *, strictly: bool = False) -> Optional[str]:
    return _get_event_attribute(
        attribute="id",
        name="pre checkout query identifier",
        event=event,
        types=(PreCheckoutQuery,),
        strictly=strictly
    )


def _get_event_attribute(
    attribute,
    name: str,
    event: Event,
    types: Iterable[type],
    *,
    strictly: bool
) -> Any:
    if type(event) in types:
        return getattr(event, attribute)
    elif strictly:
        raise InvalidEventError(f"No {name} in event" + " {event!r}!", event=event)



from typing import Union

from telebox.bot.types.types.inline_query_result_article import (
    InlineQueryResultArticle
)
from telebox.bot.types.types.inline_query_result_photo import (
    InlineQueryResultPhoto
)
from telebox.bot.types.types.inline_query_result_gif import (
    InlineQueryResultGif
)
from telebox.bot.types.types.inline_query_result_mpeg4_gif import (
    InlineQueryResultMpeg4Gif
)
from telebox.bot.types.types.inline_query_result_video import (
    InlineQueryResultVideo
)
from telebox.bot.types.types.inline_query_result_audio import (
    InlineQueryResultAudio
)
from telebox.bot.types.types.inline_query_result_voice import (
    InlineQueryResultVoice
)
from telebox.bot.types.types.inline_query_result_document import (
    InlineQueryResultDocument
)
from telebox.bot.types.types.inline_query_result_location import (
    InlineQueryResultLocation
)
from telebox.bot.types.types.inline_query_result_venue import (
    InlineQueryResultVenue
)
from telebox.bot.types.types.inline_query_result_contact import (
    InlineQueryResultContact
)
from telebox.bot.types.types.inline_query_result_game import (
    InlineQueryResultGame
)
from telebox.bot.types.types.inline_query_result_cached_photo import (
    InlineQueryResultCachedPhoto
)
from telebox.bot.types.types.inline_query_result_cached_gif import (
    InlineQueryResultCachedGif
)
from telebox.bot.types.types.inline_query_result_cached_mpeg4_gif import (
    InlineQueryResultCachedMpeg4Gif
)
from telebox.bot.types.types.inline_query_result_cached_sticker import (
    InlineQueryResultCachedSticker
)
from telebox.bot.types.types.inline_query_result_cached_document import (
    InlineQueryResultCachedDocument
)
from telebox.bot.types.types.inline_query_result_cached_video import (
    InlineQueryResultCachedVideo
)
from telebox.bot.types.types.inline_query_result_cached_voice import (
    InlineQueryResultCachedVoice
)
from telebox.bot.types.types.inline_query_result_cached_audio import (
    InlineQueryResultCachedAudio
)


InlineQueryResult = Union[InlineQueryResultArticle,
                          InlineQueryResultPhoto,
                          InlineQueryResultGif,
                          InlineQueryResultMpeg4Gif,
                          InlineQueryResultVideo,
                          InlineQueryResultAudio,
                          InlineQueryResultVoice,
                          InlineQueryResultDocument,
                          InlineQueryResultLocation,
                          InlineQueryResultVenue,
                          InlineQueryResultContact,
                          InlineQueryResultGame,
                          InlineQueryResultCachedPhoto,
                          InlineQueryResultCachedGif,
                          InlineQueryResultCachedMpeg4Gif,
                          InlineQueryResultCachedSticker,
                          InlineQueryResultCachedDocument,
                          InlineQueryResultCachedVideo,
                          InlineQueryResultCachedVoice,
                          InlineQueryResultCachedAudio]



from abc import ABC, abstractmethod
from typing import Optional


class AbstractStateStorage(ABC):

    @abstractmethod
    def save_states(
        self,
        states: list[str],
        *,
        chat_id: int,
        user_id: Optional[int] = None
    ) -> None:
        pass

    @abstractmethod
    def load_states(self, *, chat_id: int, user_id: Optional[int] = None) -> list[str]:
        pass



from typing import Union, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message
from telebox.bot.consts import message_entity_types


class TextMentionFilterCache(AbstractEventFilterCache):

    def create(self, event: Union[Message, MediaGroup]) -> set[str]:
        mentions = set()
        messages = event.messages if isinstance(event, MediaGroup) else [event]

        for message in messages:
            for entity in message.get_entities():
                if entity.type == message_entity_types.TEXT_MENTION:
                    mentions.add(
                        message.get_entity_text(entity).lower()
                    )

        return mentions


class TextMentionFilter(AbstractEventFilter):

    def __init__(self, mentions: Iterable[str], cache: TextMentionFilterCache):
        self._mentions = {i.lower() for i in mentions}
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP
        }

    def check_event(self, event: Union[Message, MediaGroup]) -> bool:
        mentions = self._cache.get(event)

        if self._mentions:
            return any(i in self._mentions for i in mentions)

        return bool(mentions)


class TextMentionFilterFactory(AbstractEventFilterFactory):

    def __init__(self):
        self._cache = TextMentionFilterCache()

    def get_filter(self, *mentions: str) -> TextMentionFilter:
        return TextMentionFilter(mentions, self._cache)



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.consts import menu_button_types
from telebox.bot.types.types.web_app_info import WebAppInfo


@dataclass
class MenuButtonWebApp(Type):
    text: str
    web_app: WebAppInfo
    type: str = menu_button_types.WEB_APP



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.photo_size import PhotoSize
from telebox.bot.types.types.file import File
from telebox.bot.types.types.mask_position import MaskPosition


@dataclass
class Sticker(Type):
    file_id: str
    file_unique_id: str
    type: str
    width: int
    height: int
    is_animated: bool
    is_video: bool
    thumb: Optional[PhotoSize] = None
    emoji: Optional[str] = None
    set_name: Optional[str] = None
    premium_animation: Optional[File] = None
    mask_position: Optional[MaskPosition] = None
    custom_emoji_id: Optional[str] = None
    file_size: Optional[int] = None



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type


@dataclass
class InputLocationMessageContent(Type):
    latitude: float
    longitude: float
    horizontal_accuracy: Optional[float] = None
    live_period: Optional[int] = None
    heading: Optional[int] = None
    proximity_alert_radius: Optional[int] = None



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type


@dataclass
class ResponseParameters(Type):
    migrate_to_chat_id: Optional[int] = None
    retry_after: Optional[int] = None



from enum import Enum


class EventType(Enum):
    MESSAGE = "message"
    EDITED_MESSAGE = "edited_message"
    CHANNEL_POST = "channel_post"
    EDITED_CHANNEL_POST = "edited_channel_post"
    MEDIA_GROUP = "media_group"
    CHANNEL_MEDIA_GROUP = "channel_media_group"
    INLINE_QUERY = "inline_query"
    CHOSEN_INLINE_RESULT = "chosen_inline_result"
    CALLBACK_QUERY = "callback_query"
    SHIPPING_QUERY = "shipping_query"
    PRE_CHECKOUT_QUERY = "pre_checkout_query"
    POLL = "poll"
    POLL_ANSWER = "poll_answer"
    MY_CHAT_MEMBER = "my_chat_member"
    CHAT_MEMBER = "chat_member"
    CHAT_JOIN_REQUEST = "chat_join_request"



import time
from queue import Queue, SimpleQueue
from typing import Optional
from threading import Lock
from dataclasses import dataclass

from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.typing import Event as Event_
from telebox.dispatcher.utils.events import get_event_chat_id, get_event_user_id


_DELAY_SECS = 0.05


@dataclass
class Event:
    event: Event_
    event_type: EventType
    chat_id: Optional[int] = None
    user_id: Optional[int] = None


class EventQueue:

    def __init__(self):
        self._events = Queue()
        self._chat_events: dict[int, SimpleQueue] = {}
        self._chat_event_lock = Lock()

    def add_event(
        self,
        event: Event_,
        event_type: EventType
    ) -> None:
        chat_id = get_event_chat_id(event)
        user_id = get_event_user_id(event)
        event = Event(
            event=event,
            event_type=event_type,
            chat_id=chat_id,
            user_id=user_id
        )

        with self._chat_event_lock:
            if chat_id in self._chat_events:
                self._chat_events[chat_id].put(event, block=False)
            else:
                if chat_id is not None:
                    self._chat_events[chat_id] = SimpleQueue()

                self._events.put(event, block=False)

    def get_event(self) -> Event:
        return self._events.get()

    def set_event_as_processed(self, event: Event) -> None:
        try:
            with self._chat_event_lock:
                if event.chat_id in self._chat_events:
                    chat_events = self._chat_events[event.chat_id]

                    if not chat_events.empty():
                        next_event = chat_events.get(block=False)
                        self._events.put(next_event, block=False)

                    if chat_events.empty():
                        del self._chat_events[event.chat_id]
        finally:
            self._events.task_done()

    def wait_events(self) -> None:
        while self._chat_events:
            time.sleep(_DELAY_SECS)

        self._events.join()



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.types.types.user import User
from telebox.bot.types.types.shipping_address import ShippingAddress


@dataclass
class ShippingQuery(Type):
    id: str
    from_: User
    invoice_payload: str
    shipping_address: ShippingAddress

    @property
    def user_id(self) -> int:
        return self.from_.id



from dataclasses import dataclass
from typing import Literal, Optional

from telebox.bot.types.type import Type


@dataclass
class ReplyKeyboardRemove(Type):
    remove_keyboard: Literal[True] = True
    selective: Optional[bool] = None



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.consts import passport_element_error_sources


@dataclass
class PassportElementErrorDataField(Type):
    type: str
    field_name: str
    data_hash: str
    message: str
    source: str = passport_element_error_sources.DATA



from typing import Union, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message


class TextFilterCache(AbstractEventFilterCache):

    def create(self, event: Union[Message, MediaGroup]) -> list[str]:
        texts = []
        messages = event.messages if isinstance(event, MediaGroup) else [event]

        for i in messages:
            text = i.get_text()

            if text is not None:
                texts.append(text)

        return texts


class TextFilter(AbstractEventFilter):

    def __init__(
        self,
        texts: Iterable[str],
        full_match: bool,
        ignore_case: bool,
        cache: TextFilterCache
    ):
        self._texts = {i.lower() for i in texts} if ignore_case else set(texts)
        self._full_match = full_match
        self._ignore_case = ignore_case
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP
        }

    def check_event(self, event: Union[Message, MediaGroup]) -> bool:
        texts = self._cache.get(event)

        if not self._texts:
            return bool(texts)

        for text in texts:
            if self._ignore_case:
                text = text.lower()

            if self._full_match:
                if text in self._texts:
                    return True
            else:
                if any(i in text for i in self._texts):
                    return True

        return False


class TextFilterFactory(AbstractEventFilterFactory):

    def __init__(self):
        self._cache = TextFilterCache()

    def get_filter(
        self,
        *texts: str,
        full_match: bool = True,
        ignore_case: bool = False
    ) -> TextFilter:
        return TextFilter(texts, full_match, ignore_case, self._cache)



from typing import Optional

from telebox.context.vars import event_context
from telebox.dispatcher.utils import events as event_utils


def get_event_chat_id(*, strictly: bool = True) -> Optional[int]:
    return event_utils.get_event_chat_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_user_id(*, strictly: bool = True) -> Optional[int]:
    return event_utils.get_event_user_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_message_topic_id(*, strictly: bool = True) -> Optional[int]:
    return event_utils.get_event_message_topic_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_sender_chat_id(*, strictly: bool = True) -> Optional[int]:
    return event_utils.get_event_sender_chat_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_message_id(*, strictly: bool = True) -> Optional[int]:
    return event_utils.get_event_message_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_callback_query_id(*, strictly: bool = True) -> str:
    return event_utils.get_event_callback_query_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_inline_query_id(*, strictly: bool = True) -> str:
    return event_utils.get_event_inline_query_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_shipping_query_id(*, strictly: bool = True) -> str:
    return event_utils.get_event_shipping_query_id(
        event_context.get(),
        strictly=strictly
    )


def get_event_pre_checkout_query_id(*, strictly: bool = True) -> str:
    return event_utils.get_event_pre_checkout_query_id(
        event_context.get(),
        strictly=strictly
    )



from .dispatcher import Dispatcher, Event
from .enums import EventType, MediaGroupContentType
from .handlers import AbstractEventHandler, AbstractErrorHandler
from .filters import (
    AbstractEventFilterFactory,
    AbstractEventFilter,
    AbstractEventBaseFilter,
    AbstractEventFilterCache,
    AbstractErrorFilterFactory,
    AbstractErrorFilter,
    AbstractErrorBaseFilter,
    AbstractErrorFilterCache
)
from .middlewares import Middleware
from .utils import MediaGroup, RateLimit, Limit, Router, get_event_chat_id, get_event_user_id


__all__ = [
    "Dispatcher",
    "Event",
    "EventType",
    "MediaGroupContentType",
    "AbstractEventHandler",
    "AbstractErrorHandler",
    "AbstractEventFilterFactory",
    "AbstractEventFilter",
    "AbstractEventBaseFilter",
    "AbstractEventFilterCache",
    "AbstractErrorFilterFactory",
    "AbstractErrorFilter",
    "AbstractErrorBaseFilter",
    "AbstractErrorFilterCache",
    "Middleware",
    "MediaGroup",
    "RateLimit",
    "Limit",
    "Router",
    "get_event_chat_id",
    "get_event_user_id"
]



from dataclasses import dataclass

from telebox.bot.types.type import Type


@dataclass
class ShippingAddress(Type):
    country_code: str
    state: str
    city: str
    street_line1: str
    street_line2: str
    post_code: str



from typing import Optional, Union, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message
from telebox.bot.consts import message_entity_types


class CommandFilterCache(AbstractEventFilterCache):

    def create(self, event: Union[Message, MediaGroup]) -> Optional[str]:
        message = event.messages[0] if isinstance(event, MediaGroup) else event

        for i in message.get_entities():
            if (i.offset == 0) and (i.type == message_entity_types.BOT_COMMAND):
                return message.get_entity_text(i)


class CommandFilter(AbstractEventFilter):

    def __init__(
        self,
        commands: Iterable[str],
        username: str,
        ignore_case: bool,
        cache: CommandFilterCache
    ):
        self._commands = set()
        username = username.lower()

        for i in commands:
            if not i.startswith("/"):
                i = f"/{i}"

            if ignore_case:
                i = i.lower()

            self._commands.update((i, f"{i}@{username}"))

        self._ignore_case = ignore_case
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.MEDIA_GROUP
        }

    def check_event(self, event: Union[Message, MediaGroup]) -> bool:
        command = self._cache.get(event)

        if command is not None:
            if not self._commands:
                return True

            if self._ignore_case:
                command = command.lower()
            elif "@" in command:
                command, username = command.split("@", 1)
                command = f"{command}@{username.lower()}"

            return command in self._commands

        return False


class CommandFilterFactory(AbstractEventFilterFactory):

    def __init__(self, username: str):
        self._username = username
        self._cache = CommandFilterCache()

    def get_filter(self, *commands: str, ignore_case: bool = True) -> CommandFilter:
        return CommandFilter(commands, self._username, ignore_case, self._cache)



from .none import NoneErrorFilter
from .re_text import RETextErrorFilter
from .text import TextErrorFilter
from .type import TypeErrorFilter


__all__ = [
    "NoneErrorFilter",
    "TypeErrorFilter",
    "TextErrorFilter",
    "RETextErrorFilter"
]



from typing import Optional

from telebox.state_machine.state import State
from telebox.state_machine.storages.storage import AbstractStateStorage
from telebox.state_machine.magazine import StateMagazine
from telebox.state_machine.errors import (
    StateNameExistsError,
    StateNotFoundError
)


class StateManager:

    def __init__(self, initial_state: State, storage: AbstractStateStorage):
        self._initial_state = initial_state
        self._storage = storage
        self._states: dict[str, State] = {}
        self.add_state(initial_state)

    @property
    def initial_state(self) -> State:
        return self._initial_state

    @property
    def states(self) -> set[State]:
        return set(self._states.values())

    def add_state(self, state: State) -> None:
        if state not in self.states:
            if state.name in self._states:
                raise StateNameExistsError(
                    "State name {state_name!r} already exists!",
                    state_name=state.name
                )

            self._states[state.name] = state

    def check_state(self, state: State) -> bool:
        return state in self.states

    def get_state(self, name: str) -> State:
        try:
            return self._states[name]
        except KeyError:
            raise StateNotFoundError(
                "State with name {state_name!r} not found!",
                state_name=name
            ) from None

    def load_magazine(
        self,
        *,
        chat_id: int,
        user_id: Optional[int] = None
    ) -> StateMagazine:
        states = self._storage.load_states(chat_id=chat_id, user_id=user_id)

        if not states:
            states = [str(self._initial_state)]

        return StateMagazine(states)

    def save_magazine(
        self,
        magazine: StateMagazine,
        *,
        chat_id: int,
        user_id: Optional[int] = None
    ) -> None:
        self._storage.save_states(list(magazine), chat_id=chat_id, user_id=user_id)



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.keyboard_button_poll_type import KeyboardButtonPollType
from telebox.bot.types.types.web_app_info import WebAppInfo


@dataclass
class KeyboardButton(Type):
    text: str
    request_contact: Optional[bool] = None
    request_location: Optional[bool] = None
    request_poll: Optional[KeyboardButtonPollType] = None
    web_app: Optional[WebAppInfo] = None



from typing import Optional

from telebox.state_machine.state import State
from telebox.dispatcher.handlers.event import AbstractEventHandler
from telebox.state_machine.errors import (
    TransitionExistsError,
    DestinationStateNotFoundError
)


TransitionDict = dict[
    State,
    dict[
        AbstractEventHandler,
        dict[
            Optional[str],
            State
        ]
    ]
]


class TransitionScheme:

    def __init__(self):
        self._transitions: TransitionDict = {}

    def add_transition(
        self,
        source_state: State,
        destination_state: State,
        handler: AbstractEventHandler,
        direction: Optional[str] = None
    ) -> None:
        if self._check_destination_state(source_state, handler, direction):
            raise TransitionExistsError(
                "Transition already exists ({source_state=}, {handler=}, {direction=})!",
                source_state=source_state,
                handler=handler,
                direction=direction
            )

        source_states = self._transitions.setdefault(source_state, {})
        handlers = source_states.setdefault(handler, {})
        handlers[direction] = destination_state

    def check_transition(
        self,
        source_state: State,
        destination_state: State,
        handler: AbstractEventHandler,
        direction: Optional[str] = None
    ) -> bool:
        try:
            return self._transitions[source_state][handler][direction] is destination_state
        except KeyError:
            return False

    def get_destination_state(
        self,
        source_state: State,
        handler: AbstractEventHandler,
        direction: Optional[str] = None
    ) -> State:
        try:
            return self._transitions[source_state][handler][direction]
        except KeyError:
            raise DestinationStateNotFoundError(
                "Destination state not found ({source_state=}, {handler=}, {direction=})!",
                source_state=source_state,
                handler=handler,
                direction=direction
            ) from None

    def _check_destination_state(
        self,
        source_state: State,
        handler: AbstractEventHandler,
        direction: Optional[str] = None
    ) -> bool:
        try:
            self._transitions[source_state][handler][direction]
        except KeyError:
            return False

        return True



import logging
from typing import Callable, Any, Optional, Union, NoReturn
from dataclasses import dataclass
from threading import Thread, Lock
from queue import Queue
import time
import uuid

from telebox.utils.thread_pool import ThreadPool
from telebox.utils.task_executor.errors import TaskExecutorError, PendingTaskNotFoundError


logger = logging.getLogger(__name__)
_PENDING_TASK_PROCESSING_DELAY_SECS = 0.1
_TASK_WAITING_DELAY_SECS = 0.1


@dataclass
class Task:
    id: str
    task: Callable
    args: tuple
    kwargs: dict[str, Any]


class TaskExecutor:

    def __init__(self, threads: int, *, allow_pending_tasks: bool = False):
        self._thread_pool = ThreadPool(threads, self._run_task_processing, with_barrier=True)
        self._allow_pending_tasks = allow_pending_tasks
        self._pending_task_processing_thread: Optional[Thread] = None
        self._tasks = Queue()
        self._pending_tasks: list[tuple[Task, float]] = []
        self._pending_task_lock = Lock()

    def __enter__(self):
        self.start_tasks()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.wait_tasks()

    def add_task(
        self,
        task: Callable,
        delay_secs: Union[int, float, None] = None,
        args: tuple = (),
        kwargs: Optional[dict[str, Any]] = None
    ) -> str:
        task_id = str(uuid.uuid4())
        task = Task(
            id=task_id,
            task=task,
            args=args,
            kwargs=kwargs or {}
        )

        if delay_secs is not None:
            if not self._allow_pending_tasks:
                raise TaskExecutorError("Pending tasks are not allowed!")

            if delay_secs < 0:
                raise ValueError("Delay seconds cannot be negative!")

            with self._pending_task_lock:
                logger.debug("Pending task added to queue: %r, delay_secs=%r.", task, delay_secs)
                self._pending_tasks.append((task, time.monotonic() + delay_secs))
                self._pending_tasks.sort(key=lambda i: i[1], reverse=True)
        else:
            logger.debug("Task added to queue: %r.", task)
            self._tasks.put(task, block=False)

        return task_id

    def remove_pending_task(self, id_: str) -> None:
        with self._pending_task_lock:
            for index, task_items in enumerate(self._pending_tasks):
                if task_items[0].id == id_:
                    self._pending_tasks.pop(index)
                    break
            else:
                raise PendingTaskNotFoundError("Pending task with ID {id!r} not found!", id=id_)

    def start_tasks(self) -> None:
        logger.debug("Tasks is starting...")

        if self._allow_pending_tasks:
            self._pending_task_processing_thread = Thread(
                target=self._run_pending_task_processing,
                daemon=True
            )
            self._pending_task_processing_thread.start()

        self._thread_pool.start()
        logger.info("Tasks started.")

    def wait_tasks(self) -> None:
        logger.info("Finishing tasks...")

        if self._allow_pending_tasks:
            while self._pending_tasks:
                time.sleep(_TASK_WAITING_DELAY_SECS)

            self._pending_task_processing_thread = None

        self._tasks.join()
        logger.info("Tasks finished.")

    def _run_task_processing(self) -> NoReturn:
        while True:
            task = self._tasks.get()

            # noinspection PyBroadException
            try:
                logger.debug("Task processing started: %r.", task)
                task.task(*task.args, **task.kwargs)
            except Exception:
                logger.exception("An error occurred while processing a task!")
            finally:
                self._tasks.task_done()
                logger.debug("Task processing finished: %r.", task)

    def _run_pending_task_processing(self) -> NoReturn:
        while True:
            with self._pending_task_lock:
                if self._pending_tasks and (time.monotonic() > self._pending_tasks[-1][1]):
                    task, _ = self._pending_tasks.pop()
                    self._tasks.put(task, block=False)
                    continue

            time.sleep(_PENDING_TASK_PROCESSING_DELAY_SECS)



from typing import Optional, Union, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message
from telebox.bot.types.types.callback_query import CallbackQuery
from telebox.state_machine.state import State
from telebox.state_machine.machine import StateMachine


class ChatStateFilterCache(AbstractEventFilterCache):

    def __init__(self, machine: StateMachine):
        self._machine = machine

    def create(self, event: Union[Message, MediaGroup, CallbackQuery]) -> Optional[State]:
        if event.chat_id is not None:
            return self._machine.get_state(chat_id=event.chat_id)


class ChatStateFilter(AbstractEventFilter):

    def __init__(self, states: Iterable[State], cache: ChatStateFilterCache):
        if not states:
            raise ValueError("No states!")

        self._states = set(states)
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP,
            EventType.CALLBACK_QUERY
        }

    def check_event(self, event: Union[Message, MediaGroup, CallbackQuery]) -> bool:
        return self._cache.get(event) in self._states


class ChatStateFilterFactory(AbstractEventFilterFactory):

    def __init__(self, machine: StateMachine):
        self._cache = ChatStateFilterCache(machine)

    def get_filter(self, *states: State) -> ChatStateFilter:
        return ChatStateFilter(states, self._cache)



TYPING = "typing"
UPLOAD_PHOTO = "upload_photo"
RECORD_VIDEO = "record_video"
UPLOAD_VIDEO = "upload_video"
RECORD_VOICE = "record_voice"
UPLOAD_VOICE = "upload_voice"
UPLOAD_DOCUMENT = "upload_document"
CHOOSE_STICKER = "choose_sticker"
FIND_LOCATION = "find_location"
RECORD_VIDEO_NOTE = "record_video_note"
UPLOAD_VIDEO_NOTE = "upload_video_note"



from typing import Union

from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message
from telebox.bot.types.types.callback_query import CallbackQuery
from telebox.bot.types.types.chat_member_updated import ChatMemberUpdated
from telebox.bot.types.types.chat_join_request import ChatJoinRequest


class ChatFilter(AbstractEventFilter):

    def __init__(self, *ids: int):
        self._ids = set(ids)

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP,
            EventType.CALLBACK_QUERY,
            EventType.MY_CHAT_MEMBER,
            EventType.CHAT_MEMBER,
            EventType.CHAT_JOIN_REQUEST
        }

    def check_event(
        self,
        event: Union[Message,
                     MediaGroup,
                     CallbackQuery,
                     ChatMemberUpdated,
                     ChatJoinRequest]
    ) -> bool:
        chat_id = event.chat_id

        if self._ids:
            return chat_id in self._ids

        return chat_id is not None



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

from telebox.errors import TeleboxError
if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event


@dataclass
class DispatcherError(TeleboxError):
    """Class for dispatcher errors."""


@dataclass
class InvalidEventError(DispatcherError):
    event: Event



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type


@dataclass
class InputVenueMessageContent(Type):
    latitude: float
    longitude: float
    title: str
    address: str
    foursquare_id: Optional[str] = None
    foursquare_type: Optional[str] = None
    google_place_id: Optional[str] = None
    google_place_type: Optional[str] = None



CHANGE_INFO = "change_info"
POST_MESSAGES = "post_messages"
EDIT_MESSAGES = "edit_messages"
DELETE_MESSAGES = "delete_messages"
RESTRICT_MEMBERS = "restrict_members"
INVITE_USERS = "invite_users"
PIN_MESSAGES = "pin_messages"
PROMOTE_MEMBERS = "promote_members"
MANAGE_VIDEO_CHATS = "manage_video_chats"
ANONYMOUS = "anonymous"
MANAGE_CHAT = "manage_chat"



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Union

from telebox.bot.types.type import Type


@dataclass
class InputFile(Type):
    content: bytes = field(repr=False)
    name: Optional[str] = None


def get_input_file_by_path(path: Union[str, Path], name: Optional[str] = None) -> InputFile:
    path = Path(path)

    with path.open("rb") as stream:
        return InputFile(
            content=stream.read(),
            name=name or path.name
        )



from contextvars import ContextVar
from abc import ABC, abstractmethod


class AbstractEventFilterCache(ABC):

    def __init_subclass__(cls, **kwargs):
        cls.__current_event_context = ContextVar(f"{cls.__name__}_current_event_context")
        cls.__result_context = ContextVar(f"{cls.__name__}_result_context")

    @abstractmethod
    def create(self, event):
        pass

    def get(self, event):
        try:
            current_event = self.__current_event_context.get()
        except LookupError:
            result = self.create(event)
            self.__current_event_context.set(event)
            self.__result_context.set(result)
        else:
            if current_event is event:
                result = self.__result_context.get()
            else:
                result = self.create(event)
                self.__current_event_context.set(event)
                self.__result_context.set(result)

        return result



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.types.types.photo_size import PhotoSize


@dataclass
class UserProfilePhotos(Type):
    total_count: int
    photos: list[list[PhotoSize]]



from .callback_data import CallbackDataFilter
from .chat import ChatFilter
from .chat_type import ChatTypeFilter
from .forwarded_message import ForwardedMessageFilter
from .left_chat_member import LeftChatMemberFilter
from .media_group_content_type import MediaGroupContentTypeFilter
from .message_content_type import MessageContentTypeFilter
from .new_chat_member import NewChatMemberFilter
from .none import NoneFilter
from .reply_message import ReplyMessageFilter
from .sender_contact import SenderContactFilter
from .user import UserFilter


__all__ = [
    "CallbackDataFilter",
    "ChatFilter",
    "ChatTypeFilter",
    "ForwardedMessageFilter",
    "LeftChatMemberFilter",
    "MediaGroupContentTypeFilter",
    "MessageContentTypeFilter",
    "NewChatMemberFilter",
    "NoneFilter",
    "ReplyMessageFilter",
    "SenderContactFilter",
    "UserFilter"
]



from typing import Optional, Any

from telebox.state_machine.machine import StateMachine
from telebox.state_machine.state import State
from telebox.context.vars import event_context, event_handler_context
from telebox.context.utils import (
    get_event_chat_id,
    get_event_user_id
)


class ContextStateMachine:

    def __init__(self, machine: StateMachine):
        self._machine = machine

    def get_state(self, *, for_chat: bool = False) -> State:
        return self._machine.get_state(
            chat_id=get_event_chat_id(),
            user_id=None if for_chat else get_event_user_id()
        )

    def set_next_state(
        self,
        direction: Optional[str] = None,
        data: Any = None,
        *,
        for_chat: bool = False
    ) -> None:
        self._machine.set_next_state(
            event=event_context.get(),
            handler=event_handler_context.get(),
            direction=direction,
            data=data,
            chat_id=get_event_chat_id(),
            user_id=None if for_chat else get_event_user_id()
        )

    def set_previous_state(self, data: Any = None, *, for_chat: bool = False) -> None:
        self._machine.set_previous_state(
            event=event_context.get(),
            data=data,
            chat_id=get_event_chat_id(),
            user_id=None if for_chat else get_event_user_id()
        )

    def set_state(self, state: State, data: Any = None, *, for_chat: bool = False) -> None:
        self._machine.set_state(
            state=state,
            event=event_context.get(),
            data=data,
            chat_id=get_event_chat_id(),
            user_id=None if for_chat else get_event_user_id()
        )

    def reset_state(
        self,
        data: Any = None,
        *,
        with_exit: bool = True,
        for_chat: bool = False
    ) -> None:
        self._machine.reset_state(
            event=event_context.get(),
            data=data,
            chat_id=get_event_chat_id(),
            user_id=None if for_chat else get_event_user_id(),
            with_exit=with_exit
        )



from typing import Union

from telebox.bot.types.types.input_text_message_content import (
    InputTextMessageContent
)
from telebox.bot.types.types.input_location_message_content import (
    InputLocationMessageContent
)
from telebox.bot.types.types.input_venue_message_content import (
    InputVenueMessageContent
)
from telebox.bot.types.types.input_contact_message_content import (
    InputContactMessageContent
)
from telebox.bot.types.types.input_invoice_message_content import (
    InputInvoiceMessageContent
)


InputMessageContent = Union[InputTextMessageContent,
                            InputLocationMessageContent,
                            InputVenueMessageContent,
                            InputContactMessageContent,
                            InputInvoiceMessageContent]



from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.bot.types.types.callback_query import CallbackQuery


class CallbackDataFilter(AbstractEventFilter):

    def __init__(self, *data: str):
        self._data = set(data)

    def get_event_types(self) -> set[EventType]:
        return {EventType.CALLBACK_QUERY}

    def check_event(self, event: CallbackQuery) -> bool:
        return event.data in self._data if self._data else event.data is not None



from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.dispatcher.enums.media_group_content_type import MediaGroupContentType


class MediaGroupContentTypeFilter(AbstractEventFilter):

    def __init__(self, *types: MediaGroupContentType):
        if not types:
            raise ValueError("No media group content types!")

        self._types = set(types)

    def get_event_types(self) -> set[EventType]:
        return {EventType.MEDIA_GROUP, EventType.CHANNEL_MEDIA_GROUP}

    def check_event(self, event: MediaGroup) -> bool:
        return all(i in self._types for i in event.content_types)



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.types.types.user import User


@dataclass
class ProximityAlertTriggered(Type):
    traveler: User
    watcher: User
    distance: int



from contextvars import ContextVar
from abc import ABC, abstractmethod


class AbstractErrorFilterCache(ABC):

    def __init_subclass__(cls, **kwargs):
        cls.__current_error_event_context = ContextVar(
            f"{cls.__name__}_current_error_event_context"
        )
        cls.__result_context = ContextVar(f"{cls.__name__}_result_context")

    @abstractmethod
    def create(self, error, event):
        pass

    def get(self, error, event):
        try:
            current_error, current_event = self.__current_error_event_context.get()
        except LookupError:
            result = self.create(error, event)
            self.__current_error_event_context.set((error, event))
            self.__result_context.set(result)
        else:
            if (current_error is error) and (current_event is event):
                result = self.__result_context.get()
            else:
                result = self.create(error, event)
                self.__current_error_event_context.set((error, event))
                self.__result_context.set(result)

        return result



from telebox.dispatcher.typing import Event
from telebox.dispatcher.enums.event_type import EventType


class Middleware:

    def pre_process_event(self, event: Event, event_type: EventType) -> None:
        pass

    def process_event(self, event: Event, event_type: EventType) -> None:
        pass

    def post_process_event(self, event: Event, event_type: EventType) -> None:
        pass

    def pre_process_error(self, error: Exception, event: Event, event_type: EventType) -> None:
        pass

    def process_error(self, error: Exception, event: Event, event_type: EventType) -> None:
        pass

    def post_process_error(self, error: Exception, event: Event, event_type: EventType) -> None:
        pass



import logging
from typing import Optional, Union, NoReturn
from dataclasses import dataclass
from threading import Thread, Lock, Event as ThreadingEvent
import contextlib
import time

from requests.exceptions import Timeout as RequestTimeoutError
import cherrypy

from telebox.bot.bot import Bot
from telebox.bot.types.types.update import Update
from telebox.bot.types.types.message import Message
from telebox.dispatcher.typing import Event
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.dispatcher.utils.event_queue import EventQueue
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.handlers.event import AbstractEventHandler
from telebox.dispatcher.handlers.error import AbstractErrorHandler
from telebox.dispatcher.filters.events.filter import AbstractEventBaseFilter
from telebox.dispatcher.filters.errors.filter import AbstractErrorBaseFilter
from telebox.dispatcher.filters.events.filters.none import NoneFilter
from telebox.dispatcher.filters.errors.filters.none import NoneErrorFilter
from telebox.dispatcher.middlewares.middleware import Middleware
from telebox.dispatcher.utils.rate_limiter.rate_limiter import RateLimiter
from telebox.dispatcher.utils.rate_limiter.rate_limit import RateLimit
from telebox.dispatcher.utils.media_group_container import MediaGroupContainer
from telebox.dispatcher.utils.server_root import ServerRoot
from telebox.dispatcher.utils.router import Router
from telebox.dispatcher.errors import DispatcherError
from telebox.utils.thread_pool import ThreadPool
from telebox.utils.not_set import NotSet, NOT_SET
from telebox.dispatcher.utils.events import get_event_chat_id
from telebox.context.vars import (
    event_context,
    event_handler_context,
    error_handler_context
)


logger = logging.getLogger(__name__)
_none_filter = NoneFilter()
_none_error_filter = NoneErrorFilter()
_MEDIA_GROUP_GATHERING_DELAY_SECS = 0.1
_DROPPED_UNKNOWN_UPDATE_MESSAGE = "Update dropped because it contains an unknown content type: %r."


@dataclass
class EventHandler:
    handler: AbstractEventHandler
    filter: AbstractEventBaseFilter
    rate_limiter: Optional[RateLimiter] = None

    @property
    def rate_limit(self) -> Optional[RateLimit]:
        return self.rate_limiter.limit if self.rate_limiter is not None else None


@dataclass
class ErrorHandler:
    handler: AbstractErrorHandler
    filter: AbstractErrorBaseFilter


class Dispatcher:

    def __init__(
        self,
        bot: Bot,
        *,
        drop_over_limit_events: bool = False,
        default_rate_limit: Optional[RateLimit] = None
    ):
        self._bot = bot
        self._drop_over_limit_events = drop_over_limit_events
        self._default_rate_limit = default_rate_limit
        self._polling_is_used = False
        self._server_is_used = False
        self._events = EventQueue()
        self._thread_pool: Optional[ThreadPool] = None
        self._event_handlers: dict[EventType, list[EventHandler]] = {i: [] for i in EventType}
        self._error_handlers: list[ErrorHandler] = []
        self._middlewares: list[Middleware] = []
        self.router = Router(self)
        self._media_group_gathering_thread: Optional[Thread] = None
        self._media_group_containers: dict[str, MediaGroupContainer] = {}
        self._media_group_message_lock = Lock()
        self._polling_stopping_event = ThreadingEvent()

    def add_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.MESSAGE, filter_, rate_limit)

    def add_edited_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.EDITED_MESSAGE, filter_, rate_limit)

    def add_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.CHANNEL_POST, filter_, rate_limit)

    def add_edited_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.EDITED_CHANNEL_POST, filter_, rate_limit)

    def add_media_group_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.MEDIA_GROUP, filter_, rate_limit)

    def add_channel_media_group_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.CHANNEL_MEDIA_GROUP, filter_, rate_limit)

    def add_inline_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.INLINE_QUERY, filter_, rate_limit)

    def add_chosen_inline_result_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.CHOSEN_INLINE_RESULT, filter_, rate_limit)

    def add_callback_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.CALLBACK_QUERY, filter_, rate_limit)

    def add_shipping_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.SHIPPING_QUERY, filter_, rate_limit)

    def add_pre_checkout_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.PRE_CHECKOUT_QUERY, filter_, rate_limit)

    def add_poll_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.POLL, filter_, None)

    def add_poll_answer_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.POLL_ANSWER, filter_, None)

    def add_my_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.MY_CHAT_MEMBER, filter_, None)

    def add_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.CHAT_MEMBER, filter_, None)

    def add_chat_join_request_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.CHAT_JOIN_REQUEST, filter_, None)

    def add_error_handler(
        self,
        handler: AbstractErrorHandler,
        filter_: Optional[AbstractErrorBaseFilter] = None
    ) -> None:
        filter_ = _get_error_filter(filter_)

        self._error_handlers.append(
            ErrorHandler(
                handler=handler,
                filter=filter_
            )
        )

    def add_middleware(self, middleware: Middleware) -> None:
        self._middlewares.append(middleware)

    def check_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.MESSAGE, filter_, rate_limit)

    def check_edited_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.EDITED_MESSAGE, filter_, rate_limit)

    def check_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.CHANNEL_POST, filter_, rate_limit)

    def check_edited_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(
            handler,
            EventType.EDITED_CHANNEL_POST,
            filter_,
            rate_limit
        )

    def check_media_group_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.MEDIA_GROUP, filter_, rate_limit)

    def check_channel_media_group_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(
            handler,
            EventType.CHANNEL_MEDIA_GROUP,
            filter_,
            rate_limit
        )

    def check_inline_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.INLINE_QUERY, filter_, rate_limit)

    def check_chosen_inline_result_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(
            handler,
            EventType.CHOSEN_INLINE_RESULT,
            filter_,
            rate_limit
        )

    def check_callback_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.CALLBACK_QUERY, filter_, rate_limit)

    def check_shipping_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.SHIPPING_QUERY, filter_, rate_limit)

    def check_pre_checkout_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(
            handler,
            EventType.PRE_CHECKOUT_QUERY,
            filter_,
            rate_limit
        )

    def check_poll_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.POLL, filter_, None)

    def check_poll_answer_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.POLL_ANSWER, filter_, None)

    def check_my_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.MY_CHAT_MEMBER, filter_, None)

    def check_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.CHAT_MEMBER, filter_, None)

    def check_chat_join_request_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.CHAT_JOIN_REQUEST, filter_, None)

    def check_error_handler(
        self,
        handler: AbstractErrorHandler,
        filter_: Optional[AbstractErrorBaseFilter] = None
    ) -> bool:
        filter_ = _get_error_filter(filter_)

        for i in self._error_handlers:
            if (i.handler is handler) and (i.filter is filter_):
                return True

        return False

    def check_middleware(self, middleware: Middleware) -> bool:
        for i in self._middlewares:
            if i is middleware:
                return True

        return False

    def run_polling(
        self,
        threads: int,
        *,
        delay_secs: Union[int, float] = 0,
        error_delay_secs: Union[int, float] = 5,
        limit: Optional[int] = None,
        timeout: Optional[int] = 10,
        allowed_updates: Optional[list[str]] = None
    ) -> None:
        if self._polling_is_used:
            raise DispatcherError("Polling cannot be run twice!")

        if self._server_is_used:
            raise DispatcherError("Polling cannot be run while the server is used!")

        if delay_secs < 0:
            raise ValueError("Delay seconds cannot be negative!")

        if error_delay_secs < 0:
            raise ValueError("Error delay seconds cannot be negative!")

        self._polling_is_used = True
        offset_update_id = None
        self._run_media_group_gathering_thread()
        self._run_thread_pool(threads)
        logger.info("Polling started.")

        with contextlib.suppress(KeyboardInterrupt):
            while not self._polling_stopping_event.is_set():
                # noinspection PyBroadException
                try:
                    updates = self._bot.get_updates(
                        timeout_secs=timeout + 1 if timeout else None,
                        offset=offset_update_id,
                        limit=limit,
                        timeout=timeout,
                        allowed_updates=allowed_updates
                    )
                except RequestTimeoutError:
                    logger.error("Timeout for requesting updates has expired!")
                except Exception:
                    logger.exception("An error occurred while receiving updates!")
                    time.sleep(error_delay_secs)
                else:
                    for i in updates:
                        self._process_update(i)

                    if updates:
                        offset_update_id = updates[-1].update_id + 1

                    time.sleep(delay_secs)

        self._polling_stopping_event.clear()
        logger.info("Polling stopped.")
        self._finish_update_processing()
        self._polling_is_used = False

    def stop_polling(self) -> None:
        if not self._polling_is_used:
            raise DispatcherError("Polling not running!")

        logger.info("Polling stopping...")
        self._polling_stopping_event.set()

    def run_server(
        self,
        threads: int,
        webhook_path: Optional[str] = None,
        *,
        host: str = "0.0.0.0",
        port: int = 443,
        ssl_certificate_path: Optional[str] = None,
        ssl_private_key_path: Optional[str] = None
    ) -> None:
        if self._server_is_used:
            raise DispatcherError("Server cannot be run twice!")

        if self._polling_is_used:
            raise DispatcherError("Server cannot be run while polling is used!")

        self._server_is_used = True
        cherrypy.config.update({
            "server.socket_host": host,
            "server.socket_port": port,
            'environment': 'production',
            "engine.autoreload.on": False
        })

        if (ssl_certificate_path is not None) and (ssl_private_key_path is not None):
            cherrypy.config.update({
                "server.ssl_module": "builtin",
                "server.ssl_certificate": ssl_certificate_path,
                "server.ssl_private_key": ssl_private_key_path,
            })

        self._run_media_group_gathering_thread()
        self._run_thread_pool(threads)
        logger.info("Server started.")
        cherrypy.quickstart(
            root=ServerRoot(self._process_update),
            script_name=webhook_path.rstrip("/") if webhook_path else str()
        )
        logger.info("Server stopped.")
        self._finish_update_processing()
        self._server_is_used = False

    def stop_server(self) -> None:
        if not self._server_is_used:
            raise DispatcherError("Server not running!")

        logger.info("Server stopping...")
        cherrypy.engine.exit()

    def drop_pending_updates(
        self,
        *,
        timeout_secs: Union[int, float, None] = None,
        with_delete_webhook: bool = True
    ) -> None:
        logger.debug("Dropping pending updates...")

        if with_delete_webhook:
            self._bot.delete_webhook(timeout_secs=timeout_secs, drop_pending_updates=True)
        else:
            updates = self._bot.get_updates(
                timeout_secs=timeout_secs,
                offset=-1
            )

            if updates:
                self._bot.get_updates(
                    timeout_secs=timeout_secs,
                    offset=updates[-1].update_id + 1
                )

        logger.info("Pending updates dropped.")

    def _add_event_handler(
        self,
        handler: AbstractEventHandler,
        event_type: EventType,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        filter_ = _get_event_filter(filter_)

        if not filter_.check_event_type(event_type):
            raise DispatcherError(f"{event_type!r} is not supported by this filter!")

        rate_limit = self._get_rate_limit(rate_limit)
        rate_limiter = RateLimiter(rate_limit) if rate_limit is not None else None
        self._event_handlers[event_type].append(
            EventHandler(
                handler=handler,
                filter=filter_,
                rate_limiter=rate_limiter
            )
        )

    def _check_event_handler(
        self,
        handler: AbstractEventHandler,
        event_type: EventType,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        filter_ = _get_event_filter(filter_)
        rate_limit = self._get_rate_limit(rate_limit)

        for i in self._event_handlers[event_type]:
            if (
                (i.handler is handler)
                and (i.filter is filter_)
                and (i.rate_limit is rate_limit)
            ):
                return True

        return False

    def _get_event_handler(self, event: Event, event_type: EventType) -> Optional[EventHandler]:
        results = {}

        for i in self._event_handlers[event_type]:
            if i.filter.get_result(event, results):
                return i

    def _get_error_handler(self, error: Exception, event: Event) -> Optional[ErrorHandler]:
        results = {}

        for i in self._error_handlers:
            if i.filter.get_result(error, event, results):
                return i

    def _get_rate_limit(
        self,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> Optional[RateLimit]:
        return rate_limit if rate_limit is not NOT_SET else self._default_rate_limit

    def _run_media_group_gathering_thread(self) -> None:
        self._media_group_gathering_thread = Thread(
            target=self._run_media_group_gathering,
            daemon=True
        )
        self._media_group_gathering_thread.start()

    def _run_thread_pool(self, threads: int) -> None:
        self._thread_pool = ThreadPool(
            threads=threads,
            target=self._run_event_processing,
            args=(self._events,),
            with_barrier=True
        )
        self._thread_pool.start()

    def _process_update(self, update: Update) -> None:
        logger.debug("Update received: %r.", update)
        event = update.content

        if (
            (event is None)
            or (isinstance(event, Message) and (event.content is None))
        ):
            logger.debug(_DROPPED_UNKNOWN_UPDATE_MESSAGE, update)

            return

        event_type = EventType(update.content_type.value)

        if (
            (event_type in frozenset((EventType.MESSAGE, EventType.CHANNEL_POST)))
            and (event.media_group_id is not None)
        ):
            event: Message

            with self._media_group_message_lock:
                if event.media_group_id not in self._media_group_containers:
                    self._media_group_containers[event.media_group_id] = MediaGroupContainer(
                        event=event,
                        event_type=event_type
                    )
                else:
                    self._media_group_containers[event.media_group_id].add_event(event)

            return

        if self._check_over_limit_event(event):
            logger.debug("Event from over limit chat dropped: %r.", event)
        else:
            logger.debug("Event added to queue: %r.", event)
            self._events.add_event(event, event_type)

    def _check_over_limit_event(self, event: Event) -> bool:
        return (
            self._drop_over_limit_events
            and get_event_chat_id(event) in self._bot.over_limit_chat_ids
        )

    def _finish_update_processing(self) -> None:
        logger.info("Finishing update processing...")
        self._events.wait_events()
        self._media_group_gathering_thread = None
        self._thread_pool = None
        logger.info("Update processing finished.")

    def _run_media_group_gathering(self) -> NoReturn:
        while True:
            with self._media_group_message_lock:
                for media_group_id in tuple(self._media_group_containers):
                    if (time.monotonic() - self._media_group_containers[media_group_id].time) >= 1:
                        container = self._media_group_containers.pop(media_group_id)
                        event = MediaGroup(container.events)

                        if self._check_over_limit_event(event):
                            logger.debug("Event from over limit chat dropped: %r.", event)
                        else:
                            if container.event_type is EventType.MESSAGE:
                                event_type = EventType.MEDIA_GROUP
                            elif container.event_type is EventType.CHANNEL_POST:
                                event_type = EventType.CHANNEL_MEDIA_GROUP
                            else:
                                raise RuntimeError(
                                    f"Unknown message type {container.event_type!r}!"
                                )

                            logger.debug("Event added to queue: %r.", event)
                            self._events.add_event(event, event_type)

            time.sleep(_MEDIA_GROUP_GATHERING_DELAY_SECS)

    def _run_event_processing(self, events: EventQueue) -> NoReturn:
        while True:
            event = events.get_event()

            try:
                logger.debug("Event processing started: %r.", event)
                event_context.set(event.event)

                for i in self._middlewares:
                    i.pre_process_event(event.event, event.event_type)

                event_handler = self._get_event_handler(event.event, event.event_type)

                if event_handler is None:
                    logger.debug("No handler found for event: %r.", event)
                    continue

                event_handler_context.set(event_handler.handler)

                if (
                    (event_handler.rate_limiter is not None)
                    and event_handler.rate_limiter.process_call(event.chat_id, event.user_id)
                ):
                    continue

                for i in self._middlewares:
                    i.process_event(event.event, event.event_type)

                event_handler.handler.process_event(event.event)

                for i in self._middlewares:
                    i.post_process_event(event.event, event.event_type)
            except Exception as error:
                # noinspection PyBroadException
                try:
                    for i in self._middlewares:
                        i.pre_process_error(error, event.event, event.event_type)

                    error_handler = self._get_error_handler(error, event.event)

                    if error_handler is None:
                        raise

                    error_handler_context.set(error_handler.handler)

                    for i in self._middlewares:
                        i.process_error(error, event.event, event.event_type)

                    error_handler.handler.process_error(error, event.event)

                    for i in self._middlewares:
                        i.post_process_error(error, event.event, event.event_type)
                except Exception:
                    logger.exception("An error occurred while processing an event!")
            finally:
                events.set_event_as_processed(event)
                logger.debug("Event processing finished: %r.", event)


def _get_event_filter(
    filter_: Optional[AbstractEventBaseFilter] = None
) -> AbstractEventBaseFilter:
    return filter_ if filter_ is not None else _none_filter


def _get_error_filter(
    filter_: Optional[AbstractErrorBaseFilter] = None
) -> AbstractErrorBaseFilter:
    return filter_ if filter_ is not None else _none_error_filter



from typing import Optional, Union, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message


class TextCommandFilterCache(AbstractEventFilterCache):

    def create(self, event: Union[Message, MediaGroup]) -> Optional[str]:
        message = event.messages[0] if isinstance(event, MediaGroup) else event
        text = message.get_text()

        if text is not None:
            return text.split(" ", 1)[0]


class TextCommandFilter(AbstractEventFilter):

    def __init__(
        self,
        commands: Iterable[str],
        prefix: str,
        ignore_case: bool,
        cache: TextCommandFilterCache
    ):
        self._commands = set()

        for i in commands:
            if not i.startswith(prefix):
                i = f"{prefix}{i}"

            if ignore_case:
                i = i.lower()

            self._commands.add(i)

        self._prefix = prefix
        self._ignore_case = ignore_case
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP
        }

    def check_event(self, event: Union[Message, MediaGroup]) -> bool:
        command = self._cache.get(event)

        if (command is not None) and command.startswith(self._prefix):
            if not self._commands:
                return True

            if self._ignore_case:
                command = command.lower()

            return command in self._commands

        return False


class TextCommandFilterFactory(AbstractEventFilterFactory):

    def __init__(self, prefix: str = ""):
        self._prefix = prefix
        self._cache = TextCommandFilterCache()

    def get_filter(self, *commands: str, ignore_case: bool = True) -> TextCommandFilter:
        return TextCommandFilter(commands, self._prefix, ignore_case, self._cache)



from .factory import AbstractErrorFilterFactory
from .filter import AbstractErrorFilter, AbstractErrorBaseFilter
from .cache import AbstractErrorFilterCache
from .filters import (
    NoneErrorFilter,
    RETextErrorFilter,
    TextErrorFilter,
    TypeErrorFilter
)


__all__ = [
    "AbstractErrorFilterFactory",
    "AbstractErrorFilter",
    "AbstractErrorBaseFilter",
    "AbstractErrorFilterCache",
    "NoneErrorFilter",
    "RETextErrorFilter",
    "TextErrorFilter",
    "TypeErrorFilter"
]



from enum import Enum


class MessageContentType(Enum):
    TEXT = "text"
    ANIMATION = "animation"
    AUDIO = "audio"
    DOCUMENT = "document"
    PHOTO = "photo"
    STICKER = "sticker"
    VIDEO = "video"
    VIDEO_NOTE = "video_note"
    VOICE = "voice"
    CONTACT = "contact"
    DICE = "dice"
    GAME = "game"
    POLL = "poll"
    VENUE = "venue"
    LOCATION = "location"
    NEW_CHAT_MEMBERS = "new_chat_members"
    LEFT_CHAT_MEMBER = "left_chat_member"
    NEW_CHAT_TITLE = "new_chat_title"
    NEW_CHAT_PHOTO = "new_chat_photo"
    DELETE_CHAT_PHOTO = "delete_chat_photo"
    GROUP_CHAT_CREATED = "group_chat_created"
    SUPERGROUP_CHAT_CREATED = "supergroup_chat_created"
    CHANNEL_CHAT_CREATED = "channel_chat_created"
    MESSAGE_AUTO_DELETE_TIMER_CHANGED = "message_auto_delete_timer_changed"
    MIGRATE_TO_CHAT_ID = "migrate_to_chat_id"
    MIGRATE_FROM_CHAT_ID = "migrate_from_chat_id"
    PINNED_MESSAGE = "pinned_message"
    INVOICE = "invoice"
    SUCCESSFUL_PAYMENT = "successful_payment"
    CONNECTED_WEBSITE = "connected_website"
    PASSPORT_DATA = "passport_data"
    PROXIMITY_ALERT_TRIGGERED = "proximity_alert_triggered"
    FORUM_TOPIC_CREATED = "forum_topic_created"
    FORUM_TOPIC_CLOSED = "forum_topic_closed"
    FORUM_TOPIC_REOPENED = "forum_topic_reopened"
    VIDEO_CHAT_SCHEDULED = "video_chat_scheduled"
    VIDEO_CHAT_STARTED = "video_chat_started"
    VIDEO_CHAT_ENDED = "video_chat_ended"
    VIDEO_CHAT_PARTICIPANTS_INVITED = "video_chat_participants_invited"
    WEB_APP_DATA = "web_app_data"



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.photo_size import PhotoSize


@dataclass
class Video(Type):
    file_id: str
    file_unique_id: str
    width: int
    height: int
    duration: int
    thumb: Optional[PhotoSize] = None
    file_name: Optional[str] = None
    mime_type: Optional[str] = None
    file_size: Optional[int] = None



from typing import Union

from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message


class ReplyMessageFilter(AbstractEventFilter):

    def __init__(self, *user_ids: int):
        self._user_ids = set(user_ids)

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP
        }

    def check_event(self, event: Union[Message, MediaGroup]) -> bool:
        if event.is_reply:
            if self._user_ids:
                return event.reply_to_message.user_id in self._user_ids

            return True

        return False



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Literal

from telebox.bot.types.types.message import Message
from telebox.bot.types.types.user import User
from telebox.bot.types.types.chat import Chat
from telebox.bot.types.types.message_entity import MessageEntity
from telebox.dispatcher.enums.media_group_content_type import MediaGroupContentType


@dataclass
class MediaGroup:
    messages: list[Message]

    def __post_init__(self):
        self._content_types = set()

        for i in self.messages:
            self._content_types.add(
                MediaGroupContentType(i.content_type.value)
            )

    def __iter__(self):
        return iter(self.messages)

    def __len__(self):
        return len(self.messages)

    def __getitem__(self, item):
        return self.messages.__getitem__(item)

    @property
    def id(self) -> str:
        return self.messages[0].media_group_id

    @property
    def chat(self) -> Chat:
        return self.messages[0].chat

    @property
    def message_id(self) -> int:
        return self.messages[0].message_id

    @property
    def message_thread_id(self) -> Optional[int]:
        return self.messages[0].message_thread_id

    @property
    def message_topic_id(self) -> Optional[int]:
        return self.messages[0].message_topic_id

    @property
    def from_(self) -> Optional[User]:
        return self.messages[0].from_

    @property
    def caption(self) -> Optional[str]:
        return self.messages[0].caption

    @property
    def caption_entities(self) -> Optional[list[MessageEntity]]:
        return self.messages[0].caption_entities

    @property
    def sender_chat(self) -> Optional[Chat]:
        return self.messages[0].sender_chat

    @property
    def forward_from(self) -> Optional[User]:
        return self.messages[0].forward_from

    @property
    def forward_from_chat(self) -> Optional[Chat]:
        return self.messages[0].forward_from_chat

    @property
    def forward_signature(self) -> Optional[str]:
        return self.messages[0].forward_signature

    @property
    def forward_sender_name(self) -> Optional[str]:
        return self.messages[0].forward_sender_name

    @property
    def forward_date(self) -> Optional[datetime]:
        return self.messages[0].forward_date

    @property
    def is_topic_messages(self) -> Optional[Literal[True]]:
        return self.messages[0].is_topic_message

    @property
    def is_automatic_forward(self) -> Optional[Literal[True]]:
        return self.messages[0].is_automatic_forward

    @property
    def reply_to_message(self) -> Optional[Message]:
        return self.messages[0].reply_to_message

    @property
    def has_protected_content(self) -> Optional[Literal[True]]:
        return self.messages[0].has_protected_content

    @property
    def author_signature(self) -> Optional[str]:
        return self.messages[0].author_signature

    @property
    def content_types(self) -> set[MediaGroupContentType]:
        return self._content_types

    @property
    def date(self) -> datetime:
        return self.messages[0].date

    @property
    def chat_type(self) -> str:
        return self.messages[0].chat_type

    @property
    def chat_id(self) -> int:
        return self.messages[0].chat_id

    @property
    def sender_chat_id(self) -> Optional[int]:
        return self.messages[0].sender_chat_id

    @property
    def user_id(self) -> Optional[int]:
        return self.messages[0].user_id

    @property
    def is_forwarded(self) -> bool:
        return self.messages[0].is_forwarded

    @property
    def is_reply(self) -> bool:
        return self.messages[0].is_reply



import time

from telebox.bot.types.types.message import Message
from telebox.dispatcher.enums.event_type import EventType


class MediaGroupContainer:

    def __init__(self, event: Message, event_type: EventType):
        self.events = [event]
        self.event_type = event_type
        self.time = time.monotonic()

    def add_event(self, event: Message) -> None:
        self.events.append(event)
        self.time = time.monotonic()



MESSAGE = "message"
EDITED_MESSAGE = "edited_message"
CHANNEL_POST = "channel_post"
EDITED_CHANNEL_POST = "edited_channel_post"
INLINE_QUERY = "inline_query"
CHOSEN_INLINE_RESULT = "chosen_inline_result"
CALLBACK_QUERY = "callback_query"
SHIPPING_QUERY = "shipping_query"
PRE_CHECKOUT_QUERY = "pre_checkout_query"
POLL = "poll"
POLL_ANSWER = "poll_answer"
MY_CHAT_MEMBER = "my_chat_member"
CHAT_MEMBER = "chat_member"
CHAT_JOIN_REQUEST = "chat_join_request"



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.sticker import Sticker
from telebox.bot.types.types.photo_size import PhotoSize


@dataclass
class StickerSet(Type):
    name: str
    title: str
    sticker_type: str
    is_animated: bool
    is_video: bool
    stickers: list[Sticker]
    thumb: Optional[PhotoSize] = None



from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.bot.types.types.message import Message


class LeftChatMemberFilter(AbstractEventFilter):

    def __init__(self, *user_ids: int):
        self._user_ids = set(user_ids)

    def get_event_types(self) -> set[EventType]:
        return {EventType.MESSAGE}

    def check_event(self, event: Message) -> bool:
        user_id = event.left_chat_member.id if event.left_chat_member is not None else None

        if self._user_ids:
            return user_id in self._user_ids

        return user_id is not None



from dataclasses import dataclass
import dataclasses
from typing import Any


@dataclass
class Group:

    def __post_init__(self):
        self._items = _get_items(self)

    def __len__(self):
        return len(self._items)

    def __bool__(self):
        return bool(self._items)

    def __iter__(self):
        return iter(self._items)

    def __repr__(self):
        return f"{type(self).__name__}({self._items!r})"

    def __contains__(self, item):
        return item in self._items

    def get(self, *excluded: Any) -> set[Any]:
        return {i for i in self._items if i not in excluded}


def _get_items(group: Group) -> set[Any]:
    items = set()
    fields = {i.name for i in dataclasses.fields(group)}

    for name, value in vars(group).items():
        if name in fields:
            if isinstance(value, Group):
                items.update(
                    _get_items(value)
                )
            else:
                items.add(value)

    return items



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.shipping_address import ShippingAddress


@dataclass
class OrderInfo(Type):
    name: Optional[str] = None
    phone_number: Optional[str] = None
    email: Optional[str] = None
    shipping_address: Optional[ShippingAddress] = None



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.types.types.labeled_price import LabeledPrice


@dataclass
class ShippingOption(Type):
    id: str
    title: str
    prices: list[LabeledPrice]



from typing import Pattern, Union, Iterable

from telebox.dispatcher.filters.events.factory import AbstractEventFilterFactory
from telebox.dispatcher.filters.events.filter import AbstractEventFilter
from telebox.dispatcher.filters.events.cache import AbstractEventFilterCache
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.bot.types.types.message import Message


class RETextFilterCache(AbstractEventFilterCache):

    def create(self, event: Union[Message, MediaGroup]) -> list[str]:
        texts = []
        messages = event.messages if isinstance(event, MediaGroup) else [event]

        for i in messages:
            text = i.get_text()

            if text is not None:
                texts.append(text)

        return texts


class RETextFilter(AbstractEventFilter):

    def __init__(self, patterns: Iterable[Pattern], full_match: bool, cache: RETextFilterCache):
        if not patterns:
            raise ValueError("No patterns!")

        self._patterns = patterns
        self._full_match = full_match
        self._cache = cache

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.CHANNEL_MEDIA_GROUP
        }

    def check_event(self, event: Union[Message, MediaGroup]) -> bool:
        for text in self._cache.get(event):
            if self._full_match:
                if any(i.fullmatch(text) is not None for i in self._patterns):
                    return True
            else:
                if any(i.search(text) is not None for i in self._patterns):
                    return True

        return False


class RETextFilterFactory(AbstractEventFilterFactory):

    def __init__(self):
        self._cache = RETextFilterCache()

    def get_filter(self, *patterns: Pattern, full_match: bool = False) -> RETextFilter:
        return RETextFilter(patterns, full_match, self._cache)



from .factory import AbstractEventFilterFactory
from .filter import AbstractEventFilter, AbstractEventBaseFilter
from .cache import AbstractEventFilterCache
from .factories import (
    CallbackKeyFilterFactory,
    CashtagFilterFactory,
    ChatStateFilterFactory,
    CommandFilterFactory,
    HashtagFilterFactory,
    MentionFilterFactory,
    RETextFilterFactory,
    TextFilterFactory,
    TextCommandFilterFactory,
    TextMentionFilterFactory,
    UserStateFilterFactory
)
from .filters import (
    CallbackDataFilter,
    ChatFilter,
    ChatTypeFilter,
    ForwardedMessageFilter,
    LeftChatMemberFilter,
    MediaGroupContentTypeFilter,
    MessageContentTypeFilter,
    NewChatMemberFilter,
    NoneFilter,
    ReplyMessageFilter,
    SenderContactFilter,
    UserFilter
)


__all__ = [
    "AbstractEventFilterFactory",
    "AbstractEventFilter",
    "AbstractEventBaseFilter",
    "AbstractEventFilterCache",
    "CallbackKeyFilterFactory",
    "CashtagFilterFactory",
    "ChatStateFilterFactory",
    "CommandFilterFactory",
    "HashtagFilterFactory",
    "MentionFilterFactory",
    "RETextFilterFactory",
    "TextFilterFactory",
    "TextCommandFilterFactory",
    "TextMentionFilterFactory",
    "UserStateFilterFactory",
    "CallbackDataFilter",
    "ChatFilter",
    "ChatTypeFilter",
    "ForwardedMessageFilter",
    "LeftChatMemberFilter",
    "MediaGroupContentTypeFilter",
    "MessageContentTypeFilter",
    "NewChatMemberFilter",
    "NoneFilter",
    "ReplyMessageFilter",
    "SenderContactFilter",
    "UserFilter"
]



from .callback_key import CallbackKeyFilterFactory
from .cashtag import CashtagFilterFactory
from .chat_state import ChatStateFilterFactory
from .command import CommandFilterFactory
from .hashtag import HashtagFilterFactory
from .mention import MentionFilterFactory
from .re_text import RETextFilterFactory
from .text import TextFilterFactory
from .text_command import TextCommandFilterFactory
from .text_mention import TextMentionFilterFactory
from .user_state import UserStateFilterFactory


__all__ = [
    "CallbackKeyFilterFactory",
    "CashtagFilterFactory",
    "ChatStateFilterFactory",
    "CommandFilterFactory",
    "HashtagFilterFactory",
    "MentionFilterFactory",
    "RETextFilterFactory",
    "TextFilterFactory",
    "TextCommandFilterFactory",
    "TextMentionFilterFactory",
    "UserStateFilterFactory"
]
